#include <getopt.h>
#include <iostream.h>
#include <stdlib.h>
#include <fstream.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <rope.h>

#include "omnithread.h"

#include "abobserver2_i.hh"
#include "omniORB2/Naming.hh"
#include "abng_server_manager.hh"
#include "abpoller2_i.hh"
#include "polconfig.hh"


/* BEG Globals from abpoll.c */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern "C" {
#ifndef __STDC__
#define __STDC__ 1
#endif
#include "poller.h"
#include "socket.h"
#include "abmalloc.h"
}

#ifndef FALSE
#define FALSE (0)
#endif

#ifndef TRUE
#define TRUE (1)
#endif

#define MESSAGE_TO_SOCKET_SIZE		768
#define ALL_CONNECTIONS			-1 /* when we TX to or RX from all connections on a port, use this value */

#undef WITH_STATISTICS

typedef struct TCPPort {
    int			serial_number;
    char		serial_name[10][32];
    int			port_number;
    int			socket_counter; /* counts length of message to be sent down socket */
    unsigned char	messageToSocket[MESSAGE_TO_SOCKET_SIZE];
    struct TCPPort	*next;
} TCPPort;

#define LOOP_LIMIT	(32)

static struct	TCPPort		*tcpports = (TCPPort *) NULL;
static struct   TCPPort		*lookup[10];	/* index on per-serial line basis */

unsigned long			tcp_rcvd = 0;		   /* number of tcp packets rcvd */

unsigned long			packets_rcvd = 0;	   /* number of serial packets rcvd */
unsigned long			packets_sent = 0;	   /* number of serial packets sent */

static char			*program;

int				statistics_port_number;
int			        sightings_port_number;
int				number_of_devices;

static int			with_statistics = FALSE;

int	DebugStartup         = FALSE;
int	DebugEfficiency      = FALSE;
int	DebugSerialComm      = FALSE;
int	DebugSensors         = FALSE;
int	DebugCORBAInterface  = FALSE;
int	DebugSocketInterface = FALSE;
int     MIN_STATION          = 0;
int     MAX_STATION          = 0;

static int			with_display 	= TRUE;


/* END Globals from abpoll.c */

#include "abngpoller.hh"

#define BAUDRATE       B9600
#define _POSIX_SOURCE  1
#define FALSE          0
#define TRUE           1

struct attributes
{
     unsigned char	network_number;
     char*		device_name;
};

omni_mutex corba_msgqueue_m; // CorbaAddMessageo & CorbaRX critical section
omni_mutex abng_init_m;
omni_mutex orl_serial_m[MAX_SERIAL_LINES];

static int abng_initialised = 0;

static SocketMessage *corba_messages = (SocketMessage *) NULL;
static SocketMessage *corba_last_message = (SocketMessage *) NULL;
extern NetworkList *networks;
extern "C" void		pollAnUnknownStation(NetworkList *);

AB_Poller_V2_i* abngpoller;

CORBA::ORB_ptr  orb;
CORBA::BOA_ptr  boa;
char* serial_device;

omni_thread *ORLpoll_t[MAX_SERIAL_LINES];
omni_thread *Olivettipoll_t[MAX_SERIAL_LINES];
omni_thread *ORB_t;
omni_thread *ProcessMessages_t;
omni_thread *Efficiency_t;

int NextNumber;

char *argv1[10];
char *argv2[10];
int   argc1,argc2;

#define POL_NAME_PREFIX "poller"
#define DEFAULT_POLLER_NAME "default"
#define MAX_CONN_FAILS 1
#define ABNG_INIT_RETRY_DELAY 10

CPollerConfig PollerConfig;

void PollerExit(char *str)
{
  cerr << str;
  exit(0);
}


int 
main(int argc, char **argv)
{
  int op;
  extern char *optarg;
  char *CfgFileName;
  NetworkList		*nPtr; 
  int create_corba_intf= 0;
  NextNumber = -1;
  
  while( (op=getopt(argc,argv,"f:h")) != EOF)
    switch(op)
      {
      case 'h' :
	usage();
	break;
      case 'f' :
	PollerConfig.SetFileName( optarg );
	break;
      }

  PollerConfig.Init();
  PollerConfig.Read();
  //  PollerConfig.Dump();


  if( PollerConfig.GlobalParams["CORBAInterface"] == crope("True") )
    create_corba_intf = 1; 

  if( PollerConfig.GlobalParams["DebugStartup"] == crope("True") )
    DebugStartup = 1;
  if( PollerConfig.GlobalParams["DebugEfficiency"] == crope("True") )
    DebugEfficiency = 1;
  if( PollerConfig.GlobalParams["DebugSerialComm"] == crope("True") )
    DebugSerialComm = 1;
  if( PollerConfig.GlobalParams["DebugSensors"] == crope("True") )
    DebugSensors = 1;
  if( PollerConfig.GlobalParams["DebugCORBAInterface"] == crope("True") )
    DebugCORBAInterface = 1;
  if( PollerConfig.GlobalParams["DebugSocketInterface"] == crope("True") )
    DebugSocketInterface = 1; 
  if( PollerConfig.GlobalParams["DebugAll"] == crope("True") )
    {
      DebugStartup         = 1;
      DebugEfficiency      = 1;
      DebugSerialComm      = 1;
      DebugSensors         = 1;
      DebugCORBAInterface  = 1;
      DebugSocketInterface = 1; 
    }

  MIN_STATION = atol(PollerConfig.GlobalParams["MinStation"].c_str());
  MAX_STATION = atol(PollerConfig.GlobalParams["MaxStation"].c_str());

  
  crope pidfile = PollerConfig.GlobalParams["PidFile"];
 
  {
    if (getuid() == 0) { // root
      FILE *pidf= fopen( pidfile.c_str() , "w");
      if (pidf) {
	fprintf(pidf,"%u", (unsigned) getpid());
	fclose(pidf);
      }
    }
  }

  ORLInit();


  // This thread tries to initialize ABng
  // When it suceeds, it runs impl_is_ready
  if (create_corba_intf) {
    if ((ORB_t= new omni_thread(ORBThread, NULL)) == NULL)
      {
	cerr << "Could not start ORB thread" << endl;
	exit(0);
      }
    ORB_t->start();
  }

  // ORL sensors polling threads (one thread per serial line)

  int sl= 0;
  for (nPtr = networks; nPtr; nPtr = nPtr->next, sl++) {
    {
      if ((ORLpoll_t[sl]= new omni_thread(ORLPollThread, nPtr)) == NULL)
	{
	  cerr << "Could not start polling thread" << endl;
	exit(0);
	}
      ORLpoll_t[sl]->start();
    }
  }
    

  // Olivetti sensors polling thread
  //  if (pthread_create(Olivettipoll_t, NULL, OlivettiPollThread, NULL) != 0)
  //{
  //  cerr << "Could not start Olivetti polling thread" << endl;
  //  exit(0);
  //}

  // Message processing thread (messages in both directions - from sensors and to sensors)
  if ((ProcessMessages_t= new omni_thread(ProcessMessagesThread, NULL)) == NULL)
    {
      cerr << "Could not start message processing thread" << endl;
      exit(0);
    }
  ProcessMessages_t->start();

  if(DebugEfficiency) {
    if ((Efficiency_t= new omni_thread(EfficiencyThread, NULL)) == NULL)
      {
	cerr << "Could not start efficiency counting thread" << endl;
	exit(0);
      }  
    Efficiency_t->start();
  }

 
  for(;;) 
    sleep(2000);
}

void
ORBThread(void *arg)
{
  CosNaming::NamingContext_var	nameserv;
  CORBA::Object_var		srvman_resolved;
  CosNaming::Name		name;
  ABng_Server_Manager::Register_Manager_var	srvman;
  ABng_Server_Manager::Name	pol_name;
  char*			ns_reference_filename;
  char			ns_reference[500];
  int			c;
  int			network_nr = 0;
  attributes*		attrs;
  int                        port_option = 0;
  char *full_poller_name = NULL;
  int l;
  PollerStatus *ps = NULL;
  ProcessControlImpl *pci = NULL;
  int initialised = 0;
  int firstinit = 1;
  char *abngpoller_name = NULL;
  char * argvtmp[] = {"abngpoller","-ORBtraceLevel","400",""};
  //int argctmp = 1;
  int argctmp = 3;

  

  cout << "ORBThread started" << endl;
  orb = CORBA::ORB_init(argctmp,argvtmp,"omniORB2");
  boa = orb->BOA_init(argctmp,argvtmp,"omniORB2_BOA");
     

  // Turn off garbage collecting of unused connections
  //
  //  omniORB::idleConnectionScanPeriod( omniORB::idleIncoming, 0);
  //  omniORB::idleConnectionScanPeriod( omniORB::idleOutgoing, 0);

  abngpoller = new AB_Poller_V2_i;     
  abngpoller->_obj_is_ready(boa);
  abngpoller->name("ABNGpoller");
  CORBA::String_var p = orb->object_to_string(abngpoller);

  abngpoller->debug(DebugStartup);          // debug set in ORLInit


  if ((ns_reference_filename = getenv("NS_IOR_FILE")) == NULL) {
    cout << "Set NS_IOR_FILE variable before running poller!" << endl;
    exit(1);
  }

  fstream	nsref_file(ns_reference_filename,ios::in);
  if (nsref_file.fail()) {
    cout << "Can't open Naming Service reference file :" <<
      ns_reference_filename << endl;
    nsref_file.close();
    exit(0);
  }
  nsref_file >> ns_reference;
  nsref_file.close();

  abngpoller->name(PollerConfig.GlobalParams["Name"].c_str());
  if(*(abngpoller->name())=='\0')
    {
      cout << "Poller name missing" << endl;
      exit(0);
    }
 
  name.length(2);
  name[0].id = CORBA::string_dup("ABng");
  name[0].kind = CORBA::string_dup("ABng");
  name[1].id = CORBA::string_dup("Bootstrap@");
  name[1].kind = CORBA::string_dup("Bootstrap@");

  pol_name.length(5);
  pol_name[0] = CORBA::string_dup("ABng");
  pol_name[1] = CORBA::string_dup("ABng_Core");
  pol_name[2] = CORBA::string_dup("System");
  pol_name[3] = CORBA::string_dup("Pollers");
  pol_name[4] = abngpoller->name();

  while(!initialised) {
    if(firstinit)
      firstinit=0;
    else {
      cout << "Retrying initialisation in " << (int)ABNG_INIT_RETRY_DELAY 
      	 << "  seconds..." << endl;
      sleep(ABNG_INIT_RETRY_DELAY);
    }


          cout << "Initialising ABng CORBA interface" << endl;
     try {
      cout << "Retrieving Naming Service" << endl;
      CORBA::Object_var tmp_obj= 
	orb->string_to_object(ns_reference);
      cout << "narrow" << endl;
      nameserv = CosNaming::NamingContext::_narrow(tmp_obj);
      cout << "po narrow" << endl;
    }
    catch(CORBA::COMM_FAILURE& ex) {  
      	  cerr << "Caught system exception COMM_FAILURE, unable to contact the "
      	       << "object." << endl;
      continue;
    }
    catch(omniORB::fatalException& ex) {
      cerr << "Caught omniORB2 fatalException. This indicates a bug is caught "
	   << "within omniORB2.\nPlease send a bug report.\n"
	   << "The exception was thrown in file: " << ex.file() << "\n"
	   << "                            line: " << ex.line() << "\n"
	   << "The error message is: " << ex.errmsg() << endl;
      continue;
    }
    catch(CORBA::SystemException& e) {
      cerr << e.NP_RepositoryId() << endl;
      cerr << "Caught a system exception." << endl;
      continue;
    }
    catch (...) {
      cerr << "caught ..." << endl;
    }
    
    try {
      	  cout << "Resolving Server Manager" << endl;
      srvman_resolved = nameserv->resolve(name);
    }
    catch(CosNaming::NamingContext::NotFound& e) {
          cout << "Object Bootstrap not found in Naming Service" << endl;
      continue;
    }
    catch(CosNaming::NamingContext::CannotProceed& e) {
      	  cout << "Sorry, cannot proceed with Naming Service" << endl;
      continue;
    }
    catch(CosNaming::NamingContext::InvalidName& e) {
      	  cout << "Sorry, invalid name" << endl;
      continue;
    }
    catch(CORBA::SystemException& e) {
      	  cout << e.NP_RepositoryId() << endl;
      cout << "Caught a system exception when resolving Bootstrap" << endl;
      continue;
    }
    srvman = ABng_Server_Manager::Register_Manager::_narrow(srvman_resolved);
    
    try {
	srvman->register_server((ABng_Poller::AB_Poller_ptr)abngpoller,
				pol_name);
    }
    catch(ABng_Server_Manager::Server_Still_Active& e) {
      cout << "ServerManager exception: server still active" << endl;
      continue;
    }
    catch(ABng_Server_Manager::Invalid_Server_Name& e) {
      cout << "ServerManager exception: invalid server name" << endl;
      continue;
    }
    catch(ABng_Server_Manager::Invalid_Server_Reference& e) {
      cout << "ServerManager exception: invalid object reference" << endl;
      continue;
    }
    catch(CORBA::SystemException& e) {
      cout << "SystemException: " 
	   << e.NP_RepositoryId() << endl;
      continue;
    }
    catch(omniORB::fatalException& e) {
      cout << "omniORB::fatalException: "
	   << e.errmsg() << endl;
      continue;
    }
    
    // passing reference to standard error and storing it in file 
    fstream	reference_file("./abngpoller.ior",ios::out|ios::trunc);
    reference_file << (char*)p << endl;
    reference_file.close();
    
    l = strlen(POL_NAME_PREFIX) + 1 + strlen(abngpoller->name() ) + 1 ;
    full_poller_name = (char*) malloc(l);
    strcpy(full_poller_name,POL_NAME_PREFIX);
    strcat(full_poller_name,"_");
    strcat(full_poller_name,abngpoller->name());
    full_poller_name[l-1] = '\0';  // remove @
    
    ps = new PollerStatus();
    pci = new  ProcessControlImpl(srvman,full_poller_name,ps);
    initialised = 1;
  }
  
  ABngInitialise();
  boa->impl_is_ready(); 
  
  delete ps;
  delete pci;
  free(full_poller_name);
}


struct TCPPort * PortInList(int port)
{
  struct TCPPort      *ptr;

  ptr = tcpports;

  if( ptr == NULL )
    return NULL;

  do
    {
      if(ptr->port_number == port)
	return ptr;
      else
	ptr = ptr->next;
    }
  while(ptr != NULL);

  return NULL;
}

void
ORLInit()
{

    int			port;
    int			i,n,type_i;
    int			loops;
    int			serial_fd;
    struct TCPPort	*newp, *ptr;
    crope linedev, linedesc, type;
 
    int argc; char **argv; // REMOVE IT!!!!
    int InList = FALSE;

    for (i = 0; i < 10; i++) {
    	lookup[i] = (TCPPort *) NULL;
    }

    sightings_port_number = 0;
    number_of_devices = 10;
    newp = (TCPPort *) NULL;
    crope device;


    for(n=0 ; n < PollerConfig.SerialLines.size() ; n++)
      {
	port = atol(  PollerConfig.SerialLines[n]["Port"].c_str() );
	device = PollerConfig.SerialLines[n]["Device"];

	if ((port < 1000) || (port > 8000))
	  PollerExit("error : Port number value must be between 1000 and 80000.\n");
	
	if( (newp=PortInList(port)) != NULL )
	  {
	    // already in list
	    strncpy(newp->serial_name[newp->serial_number], device.c_str(), 
	        sizeof(newp->serial_name[newp->serial_number]));
	  }
	else
	  {
	    // add new element
	    MALLOC(newp,TCPPort,1,"abpoll.newTcpport");
	    strncpy(newp->serial_name[0], device.c_str(), 
	        sizeof(newp->serial_name[0]));
	    newp->serial_number = 0;
	    newp->socket_counter = 0;
	    newp->port_number = port;
	    newp->next = tcpports;
	    tcpports = newp;
	  }

	newp->serial_number++;
	  
      }


    // Initialisation
   
  

    for(n=0 ; n < PollerConfig.SerialLines.size() ; n++)
      {
	linedev  = PollerConfig.SerialLines[n]["Device"];
	linedesc = PollerConfig.SerialLines[n]["Description"];
	port = atol(  PollerConfig.SerialLines[n]["Port"].c_str() );
        type = PollerConfig.SerialLines[n]["Type"];

	if (type == crope("ORL"))
	  type_i = PROTO_ORL;
	else
	  if (type == crope("Olivetti"))
	    type_i = PROTO_OLIVETTI;
	  else
	    type_i = PROTO_UNKNOWN;
	
	

	cout << "Initialising " << type << " line " << linedev << " " 
	     << linedesc << endl;

	if(( serial_fd = pollerAddSerialDevice(linedev.c_str(),linedesc.c_str(),type_i)) == -1)
	  {
	    cerr << "Failed to initialise device " << linedev << endl;
	    exit(1);
	  }
	lookup[NextNumber] = PortInList(port);
	cout << "Device initialised as: " <<  serial_fd << endl;
	cout << "Network number is: " << NextNumber << endl;

	if (!(pollerInitialiseNetwork(linedev.c_str()))) 
	  {
	    cerr << "Failed to initialise network " << linedev << endl;
	    exit(1);
	  }
	

	
	if (!(socketInitialise(port))) 
	  {
	    cerr << "Failed to initialise ORL socket interface on port " 
		 << port << endl;
	    exit(1);
	  }
        
      }
   
 
}

extern "C" void		pollAnUnknownStation(NetworkList *);

void
ORLPollThread(void *attrs)
{
  NetworkList *nPtr = (NetworkList *) attrs;

  printf("Starting PollThread for serial line %d\n", (int) nPtr->number);
 
  for (;;) {
    orl_serial_m[nPtr->number].lock();
    pollerPollNetworks(nPtr);         
    /* poll all ORL stations on a network */ 
    orl_serial_m[nPtr->number].unlock();
    usleep(500);
  }
}

void
OlivettiPollThread(void *attrs)
{
    if(DebugStartup) printf("Starting OlivettiPollThread\n");
  
    for (;;) {
      //      pollerPollNetworksOlivetti(); /* poll all Olivetti stations on all networks */
    }
}


void
ProcessMessagesThread(void *attrs)
{

  for (;;) 
    {
      FromNetToSensors();
      FromSensorsToNet();
      usleep(500);      
    }
}

void
FromSensorsToNet()
{
  PollerMessage	*pPtr;
  TCPPort *ptr;


  //
  // First the ORL socket interface
  //

  while ((pPtr = pollerRX()) != (PollerMessage *) NULL) { /* serial -> socket */
    
    if (pPtr->count == 1) {
      if (DebugSerialComm) {
	printf("received 1 byte from serial network \n");   	
	continue; /* stupid packet */
      }
    }
	    
    if (DebugSerialComm) {
      unsigned int	ui;
      printf("RX %d bytes from net %d : ", (pPtr->count) - 1, pPtr->bytes[(pPtr->count) - 1]);
      for (ui = 0; ui < (pPtr->count) -1 ; ui++) {
	printf("%02x ", pPtr->bytes[ui]);
      }
      printf("\n");
    }

    //    if (pPtr->bytes[pPtr->count - 1] < 0) {
    //      printf("packet from bad serial port %d\n", pPtr->bytes[(pPtr->count) - 1] & 0xff);
    //      continue;
    //    }
    if (pPtr->bytes[pPtr->count - 1] > 9) {
      printf("packet from bad serial port %d\n", pPtr->bytes[(pPtr->count) - 1] & 0xff);
      continue;
    }

    if ((ptr = lookup[pPtr->bytes[(pPtr->count) - 1] & 0xff]) == (TCPPort *) NULL) {
      printf("packet from uninitialised serial port %d\n", pPtr->bytes[(pPtr->count) - 1] & 0xff);
      continue;
    }

    packets_rcvd++;

    if (pPtr->count + ptr->socket_counter > MESSAGE_TO_SOCKET_SIZE) {
      socketTX(ptr->port_number, ALL_CONNECTIONS, ptr->socket_counter, ptr->messageToSocket);
      ptr->socket_counter = 0;
    }

    memcpy((char *) &(ptr->messageToSocket[ptr->socket_counter]), pPtr->bytes, pPtr->count);
	    	   
    ptr->socket_counter+=pPtr->count;

    //
    // ... now the ABng CORBA interface
    //
    if (IsABngInitialised())
      CorbaSendMessage(pPtr);

    FREE(pPtr);
    pPtr = (PollerMessage *) NULL;
  }  
}

void 
CorbaSendMessage(PollerMessage *pPtr)
{
  unsigned char  message_type, length, checksum;
  static int failed_conn_attempts = 0;

  SABobserver* cur_obs;
  ABng_Poller::Simple_AB_Observer_ptr	obs_to_delete;
 
  ABng_Poller::AB_Poller::Time_Stamp	ts;
  ABng_Poller::AB_Poller::Network_Id network_nr;
  ABng_Poller::AB_Poller::Station_Id station;

  timeval current_time;

  station         = pPtr->bytes[0];
  message_type    = pPtr->bytes[1];
  length          = pPtr->bytes[2];
  network_nr      = pPtr->bytes[3+length];
  checksum        = pPtr->bytes[2+length];

  gettimeofday(&current_time,NULL);
  ts.sec = current_time.tv_sec;
  ts.usec = current_time.tv_usec;

   

  cur_obs = SABobserver::get_observer_head();
  while(cur_obs)
    {
      if (((cur_obs->get_type() != message_type) && (cur_obs->get_type() != 0))) 
	{
	  cur_obs = cur_obs->get_next();
	  continue;
	}
      try
	{
	  if (cur_obs->get_obs_type() == SABobserver::SIMPLE)
	    {
	      ABng_Poller::AB_Poller::Message mess(3+length,3+length,pPtr->bytes);
	      cur_obs->get_simple_observer()->receive(abngpoller->name(),mess);
	    }
	  else 
	    {
	      ObserverInvokeMethod(cur_obs,pPtr, &ts, station,message_type,length,
				   network_nr,checksum);
	     
	      failed_conn_attempts = 0;
	    }
	}
      catch(...)
	{
	  if (DebugCORBAInterface) 
	    cout << "Fatal exception caught when connecting to observer" << endl;
	  if( ++failed_conn_attempts == MAX_CONN_FAILS)
	    {
	      obs_to_delete = cur_obs->get_simple_observer();
	      cur_obs = cur_obs->get_next();
	      abngpoller->detach_observer_all(obs_to_delete);
	      continue;
	    }
	}
      cur_obs = cur_obs->get_next();  
    }
   

}

void 
ObserverInvokeMethod( SABobserver* cur_obs, 
		      PollerMessage *pPtr,
		      ABng_Poller::AB_Poller::Time_Stamp *ts,
		      unsigned char station,
		      unsigned char message_type,
		      unsigned char length,
		      unsigned char network_nr,
		      unsigned char checksum
		      )
{
  ABng_Poller::AB_Poller::Sighting_Code	sighting_code;
  ABng_Poller::AB_Poller::Badge_Id	badge;
  ABng_Poller::AB_Poller::Field_Id	field;
  ABng_Poller::AB_Poller::Byte		attribs,result,port,version,
    status1, status2, cycle_data,state; 
  ABng_Poller::AB_Poller::Seq_Number	seq_num = 0;
  ABng_Poller::AB_Poller::Octets	random;
  ABng_Poller::AB_Poller::Address	address;
  ABng_Poller::AB_Poller::Tones		tones;

  unsigned int j,k,l;
  int i;
  unsigned char *pb = pPtr->bytes;
  
  

  switch(message_type)
    {


      //
      // extended observers
      //

    case AUTODIAGNOSTIC_REPLY: // 0x68
      if(!  (cur_obs->get_obs_type() == SABobserver::EXTENDED) )
	break;  

      if(DebugCORBAInterface) cout << "autodiagnostic_reply" << endl ;
      result = pb[3];

      cur_obs->get_extended_observer() 
	->autodiagnostic_reply(
			       abngpoller->name(),
			       network_nr,
			       station,
			       result, 
			       *ts);
      break;

    case GET_PORT_COMMAND_REPLY: // 0x63
      if(!  (cur_obs->get_obs_type() == SABobserver::EXTENDED) )
	break;  

      if(DebugCORBAInterface) cout << "get_output_reply" << endl ;
      port = pb[3];

      cur_obs->get_extended_observer()
	->get_output_reply(abngpoller->name(),
			   network_nr,
			   station,
			   port, 
			   *ts);
	
      break;

    case SIGHT_PERIODIC:         // 0x80
    case SIGHT_SIDE_BUT:         // 0x81
    case SIGHT_MIDDLE_BUT:       // 0x82 
    case SIGHT_FIELD_PERIODIC:   // 0x84
    case SIGHT_FIELD_SIDE_BUT:   // 0x85
    case SIGHT_FIELD_MIDDLE_BUT: // 0x86
    case SIGHT_UFIELD_PERIODIC:  // 0x88
    case SIGHT_UFIELD_SIDE_BUT:  // 0x89 
    case SIGHT_UFIELD_MIDDLE_BUT:// 0x8a

     
      
      sighting_code = message_type;
      l = pb[3]; // badge message length
      badge.length(l);
      seqcpy(badge,pb+4,l);
      field.length(0);
      attribs = checksum;
      seq_num = pb[1+length];

      // this is to avoid strange badges without sequence number 
      // ??? IB    if (length > 1 + badge.length() + 1) seq_num = buf[i++];

      cur_obs->get_observer()
	->sighting(abngpoller->name(),
		   network_nr,
		   station,
		   badge,
		   sighting_code, 
		   field,
		   (ABng_Poller::AB_Poller::Byte)1, 
		   attribs, 
		   seq_num, 
		   *ts);
      break;

      if(DebugCORBAInterface) cout << "ObserverInvokeMethod: " ;

    case STATION_STATUS_REPLY: // 0x62
     
   
      if(DebugCORBAInterface) cout << "get_station_state_reply" << endl ;
      version = pb[3];
      status1 = pb[4];
      status2 = pb[5];
      cur_obs->get_observer()
	->get_station_state_reply(abngpoller->name(),
				  network_nr,
				  station,
				  (version >> 4)&0x0F
				  ,version&0x0F,
				  status1,
				  status2,
				  *ts);

      break;

    case CHALLENGE_REPLY: // 0x40
       
    

      if(DebugCORBAInterface) cout << "challenge_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      random.length(pb[4+badge.length()]);
      seqcpy(random,pb+5+badge.length(),random.length());

      cur_obs->get_observer() 
	->challenge_reply(abngpoller->name(),
			  network_nr,
			  station,
			  badge,
			  random,
			  seq_num, // ??
			  *ts);
      break;

    case SET_HOME_REPLY: // 0x41
     
     

      if(DebugCORBAInterface) cout << "set_home_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      address.length(pb[4+badge.length()]);
      seqcpy(address,pb+5+badge.length(),address.length());
      cur_obs->get_observer() 
	->set_home_reply(abngpoller->name(),
			 network_nr,
			 station, 
			 badge,
			 address,
			 seq_num, // ??
			 *ts);     
      break;

    case BEEPER_REPLY: // 0x42
         
      if(DebugCORBAInterface) cout << "sound_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      tones.length(pb[4+badge.length()]);
      seqcpy(tones,pb+5+badge.length(),tones.length());

      cur_obs->get_observer() 
	->sound_reply(abngpoller->name(),
		      network_nr,
		      station,
		      badge,
		      tones,
		      seq_num, // ??
		      *ts);
      break;

    case SET_STATE_REPLY: // 0x43
     
      if(DebugCORBAInterface) cout << "set_state_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      state = pb[4+badge.length()];
      tones.length(4);
      seqcpy(tones,pb+4+1+badge.length(),4);
      cur_obs->get_observer() 
	->set_state_reply(abngpoller->name(),
			  network_nr,
			  station,
			  badge,
			  state,
			  tones,
			  seq_num, // ?
			  *ts);      
      break;

    case DEVICE_STATUS_REPLY: // 0x44
     
      if(DebugCORBAInterface) cout << "get_state_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      version = pb[4+badge.length()];
      status1 = pb[5+badge.length()];
      cur_obs->get_observer() 
	->get_state_reply(abngpoller->name(),
			  network_nr,
			  station,
			  badge,
			  version >> 4,
			  version&0x0F,
			  status1,
			  seq_num, // ?
			  *ts);
      break;

    case GET_HOME_REPLY: // 0x47
    
      if(DebugCORBAInterface) cout << "get_home_reply" << endl ;
      badge.length(pb[3]);
      seqcpy(badge,pb+4,badge.length());
      address.length(pb[4+badge.length()]);
      seqcpy(address,pb+5+badge.length(),address.length());

      cur_obs->get_observer()
	->get_home_reply(abngpoller->name(),
			 network_nr,
			 station,
			 badge,
			 address,
			 seq_num, // ?
			 *ts);
	
      break;



    default:
      if (DebugCORBAInterface) 
	{
	  cout << "Unknown message: " << endl;
	  for (i = 0; i < (int) pPtr->count; i++) {
	    printf("%02x ", pPtr->bytes[i]);
	  }
	  printf("\n");
	}
      
      break;
    } 
}


void
FromNetToSensors()
{
  SocketMessage	*mPtr;
  TCPPort *ptr;


  for (ptr = tcpports; ptr != (TCPPort *) NULL; ptr = ptr->next) {

    if (ptr->socket_counter > 0) {
      socketTX(ptr->port_number, ALL_CONNECTIONS, ptr->socket_counter, ptr->messageToSocket);
      ptr->socket_counter = 0;
    }
	    
    /* check for an incoming message on the socket and pass it to the serial port */

    mPtr = socketRX(ptr->port_number);

    if (mPtr != (SocketMessage *) NULL) {
      if (DebugSocketInterface) {
	int		i;
	    	
	printf("RX %d bytes from %d.%d, to net/station %d/%d : ", 
	       mPtr->length, ptr->port_number, mPtr->socket, 
	       mPtr->network, mPtr->station);

	for (i = 0; i < (int) mPtr->length; i++) {
	  printf("%02x ", mPtr->body[i]);
	}
	printf("\n");
	fflush(stdout);
      }

      tcp_rcvd++;
      packets_sent++;
      orl_serial_m[mPtr->network].lock();
      
      // a dirty hack to check for an unknown station
      // before we send a command
      // it makes sensors happy (why?)
      //
      for (NetworkList *nPtr = networks; nPtr; nPtr = nPtr->next) {
	if (nPtr->number == mPtr->network) {
	  if(nPtr->protocol == PROTO_ORL)
	    pollAnUnknownStationORL( nPtr);
	  else
	    pollAnUnknownStationOlivetti( nPtr);
	  break;
	}
      }
      //
      
      int res=
	pollerTX(mPtr->network, mPtr->station, mPtr->length, mPtr->body);
      orl_serial_m[mPtr->network].unlock();
      if (DebugSocketInterface) {
	printf("pollerTX result %d\n", res);
      }
      FREE(mPtr);
      mPtr = (SocketMessage *) NULL;
    }

  }
    //
    // Form ABng CORBA interface to sensors
    //
      
    mPtr = CorbaRX();



    if (mPtr != (SocketMessage *) NULL) {
      if (DebugCORBAInterface || DebugSerialComm) {
	int		i;
	    	
	printf("RX %d bytes from ABng to net/station %d/%d : ", 
	       mPtr->length,  
	       mPtr->network, mPtr->station);

	for (i = 0; i < (int) mPtr->length; i++) {
	  printf("%02x ", mPtr->body[i]);
	}
	printf("\n");
      }
      orl_serial_m[mPtr->network].lock();
      int res=
	pollerTX(mPtr->network, mPtr->station, mPtr->length, mPtr->body);
      orl_serial_m[mPtr->network].unlock();
      
      if (DebugCORBAInterface || DebugSerialComm) {
	printf("pollerTX result %d\n", res);
      }
      FREE(mPtr);
      mPtr = (SocketMessage *) NULL;    	
    }

   
}

void usage(void)
{
  cout << "usage: "
       << "-h  help" 
       << "-f  config_file" << endl;
  exit(1);
}

void 
PollerStatus::get_server_status( ostream &ostr)
{
  struct rusage ru;
  ostr << "ABng poller resource usage:" << endl ;
  if( getrusage(RUSAGE_CHILDREN,&ru) == 0)
    {
      ostr << "user time used: " << ru.ru_utime.tv_sec << "secs "
	   << ru.ru_utime.tv_usec << " usecs" << endl;
      ostr << "system time used: " << ru.ru_stime.tv_sec << "secs "
	   << ru.ru_stime.tv_usec << " usecs" << endl;
      ostr << "block input operations: " << ru.ru_inblock << endl;
      ostr << "block output operations: " << ru.ru_oublock << endl;    
    }
}

void 
seqcpy(_CORBA_Unbounded_Sequence__Octet &o, unsigned char* str, unsigned int l)
{
  for(unsigned int i=0 ; i<l ; i++)
    o[i] = str[i];
}

//
// Message queue from ABng interface to sensors
//

SocketMessage *
CorbaRX()
{
    SocketMessage		*mPtr;

    /* return a message if one has been received */

    corba_msgqueue_m.lock();

    if (corba_messages == (SocketMessage *) NULL) {
      corba_msgqueue_m.unlock();
      return (SocketMessage *) NULL;
    }
    else {
	mPtr = corba_messages;
	corba_messages = corba_messages->next;
	corba_msgqueue_m.unlock();
	return mPtr;
    }
}


void 
CorbaAddMessage(SocketMessage *sPtr)
{
  corba_msgqueue_m.lock();
    
  if (corba_messages == (SocketMessage *) NULL) {
    corba_messages = sPtr;
    corba_last_message = sPtr;
  }
  else {
    corba_last_message->next = sPtr;
    corba_last_message = sPtr;;
  }
  corba_msgqueue_m.unlock();
}
	

void
ABngInitialise()
{
  abng_init_m.lock();
  abng_initialised = 1;
  abng_init_m.unlock();
}

int
IsABngInitialised()
{
  int i;

  abng_init_m.lock();
  i = abng_initialised;
  abng_init_m.unlock();

  return i;
}



void EfficiencyThread(void *attrs)
{
  timeval current_time,old_time;
  int efficiency;
  
  gettimeofday(&old_time,NULL);
  
  for(;;)
    {
      sleep(10);
      gettimeofday(&current_time,NULL); 
      if (current_time.tv_sec > old_time.tv_sec + 10)
	{
	  efficiency = GetPolls()/(current_time.tv_sec - old_time.tv_sec); 
	  cout << "Efficiency: " 
	       << GetStations()
	       << " stations polled every "
	       << (double) GetStations() / efficiency 
	       << " sec." << endl;
	       
	  old_time.tv_sec = current_time.tv_sec;
	  SetPolls(0);
	}
    }
}
