#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>

void usage( const char *progname) {
  fprintf(stderr,"Usage: %s -f outfile [-n minsize] [-x maxsize]\n", 
	  progname);
  exit(1);
}

int main( int argc, char **argv) {
  const char *outfname;
  int min_size= 10; // 10 KB
  int max_size= 500; // 500 KB
  int c;
  const char *progname= argv[0];
  char buf[512];
  struct stat st;
  int rsize;
  int outfd;

  while ((c= getopt( argc,argv,"f:n:x:")) != -1) 
    switch (c) {
    case 'f':
      outfname= optarg;
    break;
    case 'n':
      min_size= atoi(optarg);
      break;
    case 'x':
      max_size= atoi(optarg);
      break;
    default:
      fprintf(stderr,"Illegal option %c\n", (char)c);
      usage(progname);
      break;
    }

  if (!outfname || min_size<0 || min_size>=max_size) usage(progname);

  outfd= open( outfname, O_RDWR | O_CREAT | O_APPEND);
  if (outfd == -1) {
    fprintf(stderr,"Could not open file for writing %s\n", outfname);
    perror("");
    return -1;
  }
  while (1) {
    if (fstat(outfd, &st) == -1) {
      perror("fstat failed");
      return -1;
    }

    if (st.st_size > (long) max_size * (long) 1024)
      if (ftruncate( outfd, (long) min_size * (long) 1024) == -1) {
	perror("ftruncate failed");
	return -1;
      }
    
    rsize= read(0 /* stdin*/,
		buf,
		sizeof(buf));
    if (rsize<=0) break;
    write(outfd, buf, rsize);
  }
  if (rsize<0) {
    perror("read failed");
    return -1;
  }
  return 0;
}


