#ifndef _SABPOLLER_I_HH
#define _SABPOLLER_I_HH

#include "abng_poller.hh"
//#include "abobserver_v2.hh"
//#include "sabpoller.hh"
#include <termios.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>


#define BAUDRATE       B9600
#define _POSIX_SOURCE  1
#define FALSE          0
#define TRUE           1

class socket_manager
{
     int port;
	
   public:
			socket_manager();
     int		open_socket(int);
     int		send_message(ABng_Poller::AB_Poller::Message,ABng_Poller::AB_Poller::Network_Id,
				     ABng_Poller::AB_Poller::Station_Id);
     int		socket_maintain();
};

class clicks
{
     enum buttons
     {
	  MIDDLE,
	  SIDE
     };

     ABng_Poller::AB_Poller::Badge_Id	badge;
     int			pressed;
     timeval			timestamp;
     clicks*			next;
     buttons			button;
     ABng_Poller::AB_Poller::Message		msg;
     int			sensor_mark;

     static clicks*		click_head;
     static pthread_mutex_t	click_mutex;
     CORBA::Boolean		compare_badge_id(ABng_Poller::AB_Poller::Badge_Id);

   public:
     static CORBA::Boolean	click_counting_flag;
     static unsigned short	max_clicks;
     static unsigned short	click_timeout;
     static unsigned short	send_it(ABng_Poller::AB_Poller::Badge_Id,timeval, buttons, ABng_Poller::AB_Poller::Message,int);
     static unsigned short	get_outdated(ABng_Poller::AB_Poller::Message&,int&,timeval);
				clicks(ABng_Poller::AB_Poller::Badge_Id,timeval, buttons, ABng_Poller::AB_Poller::Message,int);
};

class SABobserver
{
   public:
     enum observer_type
     {
	  NO_OBSERVER,
	  SIMPLE,
	  STANDARD,
	  EXTENDED
     };

   private:
     ABng_Poller::AB_Observer_V2_ptr		extended_observer;
     ABng_Poller::AB_Observer_ptr		observer;
     ABng_Poller::Simple_AB_Observer_ptr	simple_observer;
     unsigned char		type;		//kind of observed messages
     observer_type		obs_type;	//kind of observer
     SABobserver*		next;
     static SABobserver*	observer_head;
     static SABobserver*	observer_tail;
   public:
     static SABobserver*	get_observer_head();
     static SABobserver*	get_observer_tail();
     static int			remove_from_list(ABng_Poller::Simple_AB_Observer_ptr,unsigned char);

				SABobserver();
				SABobserver(ABng_Poller::AB_Observer_ptr,unsigned char);
				SABobserver(ABng_Poller::Simple_AB_Observer_ptr,unsigned char);
				SABobserver(ABng_Poller::AB_Observer_V2_ptr,unsigned char);
     SABobserver*		get_next();
     void			set_next(SABobserver*);
     ABng_Poller::AB_Observer_ptr		get_observer();
     void			set_observer(ABng_Poller::AB_Observer_ptr);
     ABng_Poller::AB_Observer_V2_ptr		get_extended_observer();
     void			set_extended_observer(ABng_Poller::AB_Observer_V2_ptr);
     ABng_Poller::Simple_AB_Observer_ptr	get_simple_observer();
     void			set_simple_observer(ABng_Poller::Simple_AB_Observer_ptr);
     unsigned char		get_type();
     void			set_type(unsigned char);
     observer_type		get_obs_type();
     void			set_obs_type(observer_type);
     void			add_to_list();
};

class Simple_AB_Poller_i: public virtual ABng_Poller::_sk_Simple_AB_Poller
{
   public:
     enum NetworkType		{ORL,OLIVETTI};

   protected:
     static CORBA::Boolean	_debug;
     CORBA::Boolean		_click_counting_flag;
     char*			_name;
     unsigned short		_max_clicks;
     unsigned short		_click_timeout;
     unsigned short		_delay;
     NetworkType		_networks[255];
     char*			_serials[255];
     int			_sensor_timeout;

   public:
     int			open_device(char*);
     void			close_device(int);
     static SABobserver*	get_observer_head();
     static void		debug(CORBA::Boolean);
     static CORBA::Boolean	debug();
     
     void		set_network_type(unsigned char, NetworkType);
     NetworkType	get_network_type(unsigned char);
     void		set_serial_name(unsigned char, char*);
     char*		get_serial_name(unsigned char);
     CORBA::Boolean	click_counting_flag();
     void		click_counting_flag(CORBA::Boolean);
     char*		name();
     void		name(const char*);
     unsigned short	max_clicks();
     void		max_clicks(unsigned short);
     unsigned short	click_timeout();
     void		click_timeout(unsigned short);
     unsigned short	delay();
     void		delay(unsigned short);
     int		sensor_timeout();
     void		sensor_timeout(int);

			    Simple_AB_Poller_i();
     virtual CORBA::Boolean attach_observer(ABng_Poller::Simple_AB_Observer_ptr);
     virtual CORBA::Boolean detach_observer(ABng_Poller::Simple_AB_Observer_ptr);
     virtual CORBA::Boolean attach_type_observer(ABng_Poller::Simple_AB_Observer_ptr,CORBA::Short);
     virtual CORBA::Boolean detach_type_observer(ABng_Poller::Simple_AB_Observer_ptr,CORBA::Short);
     virtual CORBA::Boolean detach_observer_all(ABng_Poller::Simple_AB_Observer_ptr);
     virtual void issue_command(const ABng_Poller::Simple_AB_Poller::Message&);
};

#endif










